/**
 * Project:  avr_xmini_libary_test
 * Target:   AVR-Pad X-mini or Arduino Nano board
 * Created:  2020-03-28
 * Author :  M.J.Bauer  [www.mjbauer.biz]
 *
 * Version:  1.2.3  2021-03-06  (revised comments only)
 *
 * Program verifies correct connection of an LCD module to the X-mini board.
 *
 * Refer to the library header file "lib_avrXmini.h" and/or schematic diagram
 * "AVR-X-mini-Peripheral-Wiring.pdf" for peripheral connections.
 *
 * A push-button may be connected on PD2 (from PD2 to GND).
 *
 * To test UART TX function, run a terminal emulator (PuTTY) on the host PC.
 * Set Baud rate to 38400.  The serial COM port number for PuTTY can be found using
 * Windows Device Manager under "Ports (COM & LPT)".
 */

#define  F_CPU  16000000UL

#include <avr/io.h>
#include <stdlib.h>
#include "lib_avrXmini.h"

// Set this symbol FALSE if using a properly functioning Atmel X-mini or Arduino board!
#define USING_DODGY_CHINESE_NANO_CLONE  FALSE

// Function declarations...
void  TestRoutine(void);

// Global variables...
char  test_state;


int  main( void )
{
    static  unsigned long  start_LED_duty;
    static  BOOL  flipflop;

#if USING_DODGY_CHINESE_NANO_CLONE
    // Set CPU clock pre-scaler = 1 (F_CPU = F_OSC / 2)
    CLKPR = 0x80;
    CLKPR = 0x01;
#endif

    SET_BIT(DDRB, 5);      // Set PB5 as output for LED drive

    UART_init(38400);
    UART_PutString("\n");
	UART_PutString("Hello, world! \n");
	UART_PutString("AVR X-mini LCD Test routine running. \n");

    lcd_initialise();
    lcd_command(LCD_CURSOR_OFF);
    lcd_cursor_posn(0, 0);
    lcd_print_string("X-mini test v1.2");
    lcd_cursor_posn(1, 0);
    lcd_print_string("  starting...");

    TC1_initialise();
    GLOBAL_INT_ENABLE();

    while ( 1 )
    {
        if (isTaskPending_50ms())   // 50ms period expired
            ButtonScan(3);

        // If button 'A' is hit, advance to next test step...
        if (button_hit('A'))
        {
            lcd_cursor_posn(1, 0);   // Erase line 2
            lcd_print_string("                ");
            test_state++;
        }

        if (isTaskPending_500ms())  // 500ms period expired
        {
            if (flipflop)   // True every second
            {
                TestRoutine();

                SET_BIT(PORTB, 5);      // Turn LED on
                start_LED_duty = milliseconds();
            }
            flipflop = !flipflop;
        }

        if (milliseconds() >= (start_LED_duty + 100))  // 100ms duty expired
            CLEAR_BIT(PORTB, 5);   // Turn LED off
    }
}


/*
* The test routine is a state machine (sequencer) called once every second.
*/
void  TestRoutine(void)
{
    static  int   countdown;     // unit = seconds
    static  int   state_timer;   // unit = seconds
    char    buff[16];

    if (test_state == 0)    // Showing startup message for 3 seconds
    {
        if (state_timer >= 3)  // 3 seconds elapsed
        {
            countdown = 10;
            test_state = 2;
        }
    }
    else if (test_state == 1)    // Showing count-down
    {
        lcd_cursor_posn(1, 0);   // Erase line 2
        lcd_print_string("                ");
        itoa(countdown, buff, 10);
        lcd_cursor_posn(1, 6);

        if (countdown == 0)
        {
            lcd_cursor_posn(1, 6);
            lcd_print_string("BANG!");
            state_timer = 0;
            test_state = 2;
        }
        else  lcd_print_string(buff);   // Display countdown

        countdown--;
    }
    else if (test_state == 2)  // Showing message for 3 seconds
    {
        if (state_timer >= 3)  // 3 seconds elapsed
        {
            lcd_cursor_posn(1, 0);
            lcd_print_string(" Press button A ");
            test_state = 3;
        }
    }
    else if (test_state == 3)  // Waiting for button 'A' hit
    {
        ;  // Nothing to do until button hit, then step = 4
    }
    else if (test_state == 4)  // Transient state -- back to step 1
    {
        countdown = 10;
        test_state = 1;   // Repeat test #1
    }

    state_timer++;
}

