/*
 * Project:  avrbed_test_demo  (version 1.0)
 * File:     main.c
 *
 * Created:  2019-11-27
 * Author :  Mike Bauer  [www.mjbauer.biz]
 *
 * This program verifies functions in the AVR-BED library module (lib_avrbed).
 * The application also tests and demonstrates AVR-BED features.
 *
 * The test routine sequences through the following steps:
 * ```````````````````````````````````````````````````````
 *   1. Flashing 8 LEDs individually while Debug LED pulses at 1Hz
 *   2. LED pattern changes according to which button (A,B,C,D) is pressed
 *   3. ADC readings from PC1/ADC1 and PC3/ADC3 (POT) are displayed on LCD
 *   4. Key hits on 4x4 keypad and RX data from UART are displayed on LCD
 *
 *   At each step, prompts are displayed on the LCD.
 *   LCD contrast should be adjusted optimally at Step 1.
 *   Text messages are sent via the UART to verify UART-USB TX function.
 *   If using a PC terminal emulator (PuTTY), set Baud-rate to 9600.
 *
 * References:
 *   [1] "AVR-BED Development Board" (Design & construction notes)
 *   [2] "AVR-BED Library Ref Manual" (Library function usage)
 *
 */
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>
#include "lib_avrbed.h"

#define  CHASER_STEP  200  // Sets LED chaser speed (unit = ms)

// ---------- Function prototypes ---------------
void  Prepare_Step1();
void  Test1_LEDs_Flashing();
void  Test2_Buttons();
void  Test3_POT_ADC_inputs();
void  Test4_KeyPad_LCD_UART();
void  itoa_fix(unsigned val, char *dest, BYTE ndigs);
char  *strfetch(const char *progstr);

// ----------  Global variables  ----------------
BYTE   testStep;
short  button_D_hit_count;
unsigned long  startDebugLEDduty;
char  *blankline = "                ";

// Example of how to store strings in program memory (flash) instead of data SRAM.
// This requires the header file <avr/pgmspace.h> to be #included, as above.
// The function strfetch() is used to copy a string from program to data memory.
const char Step1str1[] PROGMEM = "Step 1: Verify 8 LEDs flashing (light chaser)\n";
const char Step1str2[] PROGMEM = "        Verify Debug LED pulsing at 1Hz (10% duty)\n";
const char Step1str3[] PROGMEM = "        Verify LCD working;  adjust contrast.\n";
const char Step1str4[] PROGMEM = "Next:   Press button [A] on AVR-BED \n";

const char Step3str1[] PROGMEM = "Step 3: Set potentiometer jumper to PC1/ADC1.\n";
const char Step3str2[] PROGMEM = "        Move pot from 0 to 100% and verify ADC readings on ADC1.\n";
const char Step3str3[] PROGMEM = "        (At 0 ADC = 0;  at 50% ADC = 511;  at 100% ADC = 1023) \n";
const char Step3str4[] PROGMEM = "        Set pot jumper to PC3/ADC3; verify ADC readings on ADC3.\n";
const char Step3str5[] PROGMEM = "Next:   Press button [A] or [D] on AVR-BED.\n";


int  main(void)
{
	BYTE  flipflop = 0;

	SET_BIT(DDRD, 7);   // Output to drive Debug LED
	UART_init(9600);
	TC1_initialize();
	GLOBAL_INT_ENABLE();

	lcd_initialise();       // Also sets DDRB = 0xFF (for LED drive)
	lcd_command(LCD_CURSOR_OFF);
	lcd_cursor_posn(0, 0);
	lcd_print_string("*** AVR^BED *** ");
	lcd_cursor_posn(1, 0);
	lcd_print_string("Test & Demo v1.0");

	UART_PutString("\n\n");
	UART_PutString("*** AVR^BED Test & Demo v1.0 -- MJB 2020 ***\n\n");

	_delay_ms(2000);  // Delay to show start-up screen

	Prepare_Step1();
	startDebugLEDduty = milliseconds();

 	while ( TRUE )
	{
		if (isTaskPending_50ms())   // True every 50ms
		{
			if (testStep == 4) KeypadScan();
			else  ButtonScan(4);

			if (testStep == 1)  Test1_LEDs_Flashing();
			if (testStep == 2)  Test2_Buttons();
			if (testStep == 3)  Test3_POT_ADC_inputs();
			if (testStep == 4)  Test4_KeyPad_LCD_UART();
		}

		// the Debug LED should be flashing in all tests
		if (isTaskPending_500ms())
		{
			if (flipflop)  // True once every second
			{
				SET_BIT(PORTD, 7);   // turn debug LED on
				startDebugLEDduty = milliseconds();
			}
			flipflop = ~flipflop;  // toggle
		}

		if (milliseconds() >= (startDebugLEDduty + 100))  // 100ms interval ended
			CLEAR_BIT(PORTD, 7);  // turn debug LED off
	}
}


void  Prepare_Step1()
{
	testStep = 1;

	lcd_cursor_posn(0, 0);
	lcd_print_string("Step 1: LED test");
	lcd_cursor_posn(1, 0);
	lcd_print_string("Next: press [A] ");

	UART_PutString("\n");   // new line
	UART_PutString(strfetch(Step1str1));
	UART_PutString(strfetch(Step1str2));
	UART_PutString(strfetch(Step1str3));
	UART_PutString(strfetch(Step1str4));
}


void  Test1_LEDs_Flashing()
{
	static  BYTE  chaserPattern[] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
	static  BYTE  frame = 0;
	static  BYTE  calls = 0;
	static  BYTE  direction = 0;  // 0 is Fwd, 1 is Rev.

	if (++calls == 2)
	{
		PORTB = chaserPattern[frame];  // show next frame

		if (direction == 0)  // going left to right
		{
			if (++frame == 8)
			{
				frame = 7;
				direction = 1;
			}
		}
		else  // going right to left
		{
			if (--frame == 0xFF)
			{
				frame = 0;
				direction = 0;
			}
		}
		calls = 0;
	}

	if (button_hit('A'))   // check for next step
	{
		PORTB = 0;  // 8 LEDs off
		lcd_cursor_posn(0, 0);
		lcd_print_string("Step 2: Buttons ");
		lcd_cursor_posn(1, 0);
		lcd_print_string("[A] [B] [C] [D] ");

		UART_PutString("\n");   // blank line
		UART_PutString("Step 2: Verify LED pattern changes with button pressed...\n");
		UART_PutString("Next:   Press buttons [A], [B], [C], then [D] (3 times) \n");

		button_D_hit_count = 0;
		testStep = 2;
	}
}


void  Test2_Buttons()
{
	BYTE  led_bits = 0;

	if (button_pressed('A'))  led_bits |= 0b11000000;
	if (button_pressed('B'))  led_bits |= 0b00110000;
	if (button_pressed('C'))  led_bits |= 0b00001100;
	if (button_pressed('D'))  led_bits |= 0b00000011;

	PORTB = led_bits;

	if (button_hit('D'))  button_D_hit_count++;   // check for next step

	if (button_D_hit_count >= 3)
	{
		PORTB = 0;  // 8 LEDs off
		lcd_cursor_posn(0, 0);
		lcd_print_string("Step 3: POT/ADC ");

		UART_PutString("\n");   // blank line
		UART_PutString(strfetch(Step3str1));
		UART_PutString(strfetch(Step3str2));
		UART_PutString(strfetch(Step3str3));
		UART_PutString(strfetch(Step3str4));
		UART_PutString(strfetch(Step3str5));

		button_hit('A');   // Clear button hits
		button_hit('B');
		button_hit('C');
		button_hit('D');
		testStep = 3;
	}
}


void  Test3_POT_ADC_inputs()
{
	static char  nbuff[8];
	static BYTE  count_to_4 = 0;
	int  adcReading;

	if (++count_to_4 == 4)   // Update ADC readings every 200 ms (5/sec)
	{
		count_to_4 = 0;

		adcReading = ADC_ReadInput(1);   // read ADC1 input
		itoa_fix(adcReading, nbuff, 4);
		lcd_cursor_posn(1, 0);
		lcd_print_string("A1=");
		lcd_print_string(nbuff);

		adcReading = ADC_ReadInput(3);   // read ADC3 input
		itoa_fix(adcReading, nbuff, 4);
		lcd_cursor_posn(1, 8);
		lcd_print_string("A3=");
		lcd_print_string(nbuff);
	}

	if (button_hit('A') || button_hit('D'))   // check for next step
	{
		lcd_cursor_posn(0, 0);
		lcd_print_string("Step 4: Keypad+");
		lcd_cursor_posn(1, 0);
		lcd_print_string(blankline);
		lcd_cursor_posn(1, 0);

		UART_PutString("\n");   // blank line
		UART_PutString("Step 4: Verify 4x4 keypad operation -- key hits shown on LCD \n");
		UART_PutString("        Verify PC terminal input - UART RX data shown on LCD \n");
		UART_PutString("Next:   Press [#] on keypad or reset MCU to quit.\n");
		testStep = 4;
	}
}


void  Test4_KeyPad_LCD_UART()
{
	static BYTE  cpos = 0;
	char   c;
	BOOL   hash_key_hit = FALSE;

	if (key_hit())
	{
		c = key_code();
		UART_PutByte(c);
		if ( c == '#')  hash_key_hit = TRUE;
	}
	else if (UART_RxDataAvail())  c = UART_GetByte();
	else  c = 0;

	if (c != 0)   // Got an ASCII character/code
	{
		if (cpos >= 16)   // if reached end of line...
		{
			cpos = 0;
			lcd_cursor_posn(1, 0);
			lcd_print_string(blankline);
			lcd_cursor_posn(1, 0);
		}
		lcd_write_char(c);
		cpos++;
	}

	if (hash_key_hit)  // if [#] key hit, go back to step 1
	{
		Prepare_Step1();
	}
}


/* ````````````````````````````````````````````````````````````````````````````````````
 * Convert unsigned integer value to a decimal string with a fixed number of digits.
 * Entry arg's:  val   =  unsigned integer to be converted
 *               dest  =  pointer to destination string (array name)
 *               ndigs =  number of digits in output string (1..5)
 * If the value is too big to fit into ndigs decimal places, then only the least
 * significant digits will be output.  Leading zeros (if any) are output.
 * The size declared for the output buffer (array) must be at least ndigs + 1.
 */
void  itoa_fix(unsigned val, char *dest, BYTE ndigs)
{
	short  i = ndigs - 1;   // index into output array

	if (ndigs > 5)  { ndigs = 5;  i = 4; }  // find LSD place

	dest[ndigs] = 0;  // output NUL terminator

	while (ndigs != 0)
	{
		dest[i] = '0' + val % 10;  // output ASCII digit (LSD first)
		val = val / 10;
		ndigs--;  i--;
	}
}


/* ````````````````````````````````````````````````````````````````````````````````````
 * Fetch string from (flash) program memory and copy to data memory (SRAM).
 * Entry arg:   str   =  address of source string in program memory
 * Return val:  dest  =  address of destination string in data memory
 * Notes:
 * The maximum size of the string is 80 chars, incl. NUL terminator.
 * The macro pgm_read_byte() is defined in the header file <avr/pgmspace.h>.
 * Source strings should be declared as in this example:
 * const char string_name[] PROGMEM = "Text comprising the string...";
 */
char  *strfetch(const char *str)
{
	static char  dest[80];
	char   c;
	int    n = 0;

	while (n < 79)
	{
		c = pgm_read_byte(str++);
		dest[n++] = c;
		if (c == 0)  break;   // found NUL
	}
	dest[n] = 0;   // Add NUL

	return  dest;
}
